(* TENLAB AST Module by Mehmet Kerem Turkcan *)
(* Based on the corresponding MICROC Module  *)

type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq

type expr =
    Literal of string
  | Id of string
  | TensorGet of string
  | Binop of expr * op * expr
  | Assign of string * expr
  | Call of string * expr list
  | GetTensorElement of string * string
  | Noexpr

type stmt =
    Block of stmt list
  | Expr of expr
  | BuildTensorProd of string * int list * int list * string * int list 
      * int list * string * int list
  | CreateTensor of string * string
  | Return of expr
  | Clear of string
  | Clean of string
  | If of expr * stmt * stmt
  | For of string * string * stmt
  | While of expr * stmt

type func_decl = {
    fname : string;
    formals : string list;
    locals : string list;
    body : stmt list;
  }

type program = string list * func_decl list

let rec string_of_expr = function
    Literal(l) -> l
  | Id(s) -> s ^ ".Content[0]"
  | TensorGet(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^
      (match o with
	Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
      | Equal -> "==" | Neq -> "!="
      | Less -> "<" | Leq -> "<=" | Greater -> ">" | Geq -> ">=") ^ " " ^
      string_of_expr e2
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""
  | _ -> raise (Failure ("TENLAB Error: Something impossible was observed."))

type bstmt =
    (* Direct Output of Numeric Literal *)
    Lit of string 
	(* Direct Output of String *)
  | DirectOut of string    
	(* Tensor of String *)
  | TensorGet of string  
	(* Tensor Variable of String *)
  | TempVar of string      
	(* Tensor Declaration of String *)
  | VarDeclare of string   
	(* Tensor Operation of Strings *)
  | TensorOpt of string * string * string  
    (* Direct Output of Tensor *)  
  | TensorAssign of string 
	(* Free Memory of Temporary Variable *)
  | TempMemFree of int               
	(* Get Size of Temporary Variable *)
  | TempGetSize of int                         
	(* Direct Output of String *)
  | TensorSub of string                        
    (* Call Function by Explicit Name *)
  | FunctionCall of string                     
    (* Return Values at the End of Function *)
  | ReturnValue of int                         
    (* Clear a Tensor *)
  | ClearValue of string                       
	(* Clean a Tensor from memory *)
  | CleanValue of string                       
    (* Return Values at the End of Function *)
  | VarAssign of int                           
    (* Put Right Paranthesis *)
  | Parend                                     
    (* End of Function *)
  | Curlend                                    
    (* Begin If Statement *)
  | Beginf                                     
    (* Put a Comma *)
  | Comma                                      
    (* End Line and Begin New Line *)
  | Lineend                                    
    (* Begin Curl for Blocks and Start a New Line*)
  | Curlbegin                                  
    (* Arithmetic Operations *)
  | Bin of op                                  

type prog = {
    num_globals : int;                         (* Number of global variables *)
    text : bstmt array;                        (* Code for all the functions *)
  }

(* 
Match Implicit Blocks
*)

let string_of_stmt = function
    Lit(i) -> i
  | DirectOut(i) -> i
  | TempVar(i) -> 
      "TENLAB_Tensor temp" ^ i ^ ";\nTENLAB_Tensor_create(&" ^ i ^ ");\n" 
  | TensorGet(i) -> i
  | TempMemFree(i) -> "TENLAB_Tensor_destroy(&temp" ^ 
	  string_of_int i ^ ")"
  | TempGetSize(i) -> "temp" ^ string_of_int i ^ ".cur_content_length"
  | TensorSub(i) -> i
  | VarDeclare(i) -> i
  | TensorAssign(i)  -> i ^ " = "
  | TensorOpt(i,j,k)  -> i ^ j ^ k
  | FunctionCall("input")  -> "TENLAB_Tensor_populate_from_MEX(prhs,&"
  | FunctionCall("output")  -> "TENLAB_Tensor_to_MEX(plhs,&"
  | FunctionCall("pop")  -> "TENLAB_pop_element(&"
  | FunctionCall("dequeue")  -> "TENLAB_dequeue_element(&"
  | FunctionCall("reshape")  -> "TENLAB_Tensor_reshape(&"
  | FunctionCall("print")  -> "TENLAB_Tensor_print("
  | FunctionCall("shape")  -> "TENLAB_Tensor_shape_print("
  | FunctionCall("set")  -> "TENLAB_add_element_at_specific_position(&"
  | FunctionCall("length")  -> "TENLAB_add_length(&"
  | FunctionCall("main")  -> "int main("
  | FunctionCall(i)  -> i ^ "("
  | VarAssign(i)  -> "double var" ^ string_of_int i ^ " = "
  | ReturnValue(i)  -> "return " 
  | ClearValue(i) -> 
      "TENLAB_Tensor_destroy(&" ^ i ^ ");\nTENLAB_Tensor_create(&" ^ i ^ ")"
  | CleanValue(i)  -> "TENLAB_Tensor_destroy(&" ^ i ^ ");"
  | Bin(Add)  -> "+"
  | Bin(Sub)  -> "-"
  | Bin(Mult)  -> "*"
  | Bin(Div)  -> "/"
  | Bin(Equal)  -> "=="
  | Bin(Neq)  -> "!="
  | Bin(Less)  -> "<"
  | Bin(Leq)  -> "<="
  | Bin(Greater)  -> ">"
  | Bin(Geq)  -> ">="
  | Parend  -> ")"
  | Curlend  -> "}\n"
  | Beginf  -> "if ("
  | Comma  -> ","
  | Lineend  -> ";\n"
  | Curlbegin  -> "{\n"

let string_of_vdecl id = "int " ^ id ^ ";\n"

(* 
Generate Program Text 
*)

(* 
TENLAB Preamble: 
Populate for the Language Dependencies via C Libraries and the TENLAB C Backend 
*)
let string_of_prog p =
  "#include <stdio.h>\n#include <math.h>\n#include <stdlib.h>\n" ^ 
  "#include \"tenlab_preamble.c\"\n// Stats: " ^ string_of_int p.num_globals ^
  " Script Variables;\n\n" ^
  let funca = Array.mapi
      (fun i s -> "" ^ string_of_stmt s) p.text
  in String.concat "" (Array.to_list funca)
  
(* 
Generate Program Text for MEX-Supported Programs
*)

(* 
TENLAB Preamble: 
Populate for the Language Dependencies via C Libraries and the TENLAB C Backend 
*)
let string_of_prog_mex p =
"#include <stdio.h>\n#include <math.h>\n#include <stdlib.h>\n#include <mex.h>\n" 
^ "#include \"tenlab_preamble_mex.cpp\"\n// Stats: " ^ 
string_of_int p.num_globals ^
  " Script Variables;\n\n" ^
  let funca = Array.mapi
      (fun i s -> "" ^ string_of_stmt s) p.text
  in String.concat "" (Array.to_list funca)

