/******************************************************
 *                                                    *
 * tenlab_preamble_mex.c                              *
 *                                                    *
 * Contains the MATLAB MEX code generator for TENLAB. *
 *                                                    *
 ******************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

typedef struct TENLAB_Tensor
{
	int *Shape;
	double *Content;
	size_t max_content_length;
	size_t cur_content_length;
	size_t shape_length;
} TENLAB_Tensor;

void TENLAB_Terminate() 
{
  exit(1);
}

void TENLAB_assign(TENLAB_Tensor *X, double y)
{
    if(X->cur_content_length == X->max_content_length)
    {
        int new_max_content_length = X->max_content_length + X->Shape[X->shape_length-1];
		//Different OS's may have SIZE_T_MAX instead
        if((new_max_content_length > X->max_content_length) && (new_max_content_length < SIZE_MAX / sizeof(double)))
        {
            double *new_Content = (double*) realloc(X->Content, new_max_content_length * sizeof(double));
            if(new_Content != NULL)
            {
				X->Content = new_Content;
				X->max_content_length = new_max_content_length;
            }
            else
            {
				printf("\n TENLAB Error: Memory filled during size reallocation.");
				TENLAB_Terminate();
			}
        }
        else
        {
			printf("\n TENLAB Error: Memory overflow.");
			TENLAB_Terminate();
		}
    }
    X->Content[X->cur_content_length] = (double) y;
	X->Shape[0] = X->Shape[0]+1;
    X->cur_content_length++;
}

void TENLAB_add_shape(TENLAB_Tensor *X, double y)
{
	int new_max_content_length = X->shape_length + 1;
	//Different OS's may have SIZE_T_MAX instead
	if (new_max_content_length < SIZE_MAX / sizeof(int))
	{
		int *new_Shape = (int*) realloc(X->Shape, new_max_content_length * sizeof(int));
		if(new_Shape != NULL)
		{
			X->Shape = new_Shape;
		}
		else
		{
			printf("\n TENLAB Error: Memory filled during size reallocation.");
			TENLAB_Terminate();
		}
	}
	else
	{
		printf("\n TENLAB Error: Memory overflow.");
		TENLAB_Terminate();
	}
    //if (X->shape_length==1)
    //{
    //    X->Shape[X->shape_length-1] = (int) y;
    //    X->shape_length++;
    //}
    //else
    //{
        X->shape_length++;
        X->Shape[X->shape_length-2] = (int) y;
        X->Shape[X->shape_length-1] = (int) 1;
    //}
}

void TENLAB_add_element_at_specific_position(TENLAB_Tensor *X,TENLAB_Tensor *Y, double y)
{
	if (Y->cur_content_length==X->shape_length)
	{
		int adding_index=1;
		for (int i=0;i<X->shape_length;i++)
			adding_index=adding_index * Y->Content[i];
		adding_index=(int)adding_index;
		if(X->cur_content_length < adding_index-1)
		{
			int new_max_content_length = adding_index-1;
			if((new_max_content_length > X->max_content_length) && (new_max_content_length < SIZE_MAX / sizeof(double)))
			{
				double *new_Content = (double*) realloc(X->Content, new_max_content_length * sizeof(double));
				if(new_Content != NULL)
				{
					X->Content = new_Content;
					X->max_content_length = new_max_content_length;
					for (int i=X->cur_content_length;i<=adding_index-1;i++)
						X->Content[i]=0;
					X->cur_content_length=adding_index-1;
				}
				else
				{
					printf("\n TENLAB Error: Memory filled during size reallocation.");
					TENLAB_Terminate();
				}
			}
			else if (new_max_content_length >= SIZE_MAX / sizeof(double))
			{
				//printf("\n TENLAB Error: Memory overflow.");
				//TENLAB_Terminate();
			}
		}
		X->Content[adding_index-1] = y;
	}
	else
	{
		printf("\n TENLAB Error: Assignment dimension does not match.");
		TENLAB_Terminate();
	}
}

void TENLAB_add_element_at_linear_index(TENLAB_Tensor *X,int adding_index, double y)
{
	if (X->cur_content_length==X->shape_length)
	{
		if(X->cur_content_length < adding_index-1)
		{
			int new_max_content_length = adding_index-1;
			if((new_max_content_length > X->max_content_length) && (new_max_content_length < SIZE_MAX / sizeof(double)))
			{
				double *new_Content = (double*) realloc(X->Content, new_max_content_length * sizeof(double));
				if(new_Content != NULL)
				{
					X->Content = new_Content;
					X->max_content_length = new_max_content_length;
					for (int i=X->cur_content_length;i<=adding_index-1;i++)
						X->Content[i]=0;
					X->cur_content_length=adding_index-1;
				}
				else
				{
					printf("\n TENLAB Error: Memory filled during size reallocation.");
					TENLAB_Terminate();
				}
			}
			else if (new_max_content_length >= SIZE_MAX / sizeof(double))
			{
				//printf("\n TENLAB Error: Memory overflow.");
				//TENLAB_Terminate();
			}
		}
		X->Content[adding_index-1] = y;
	}
	else
	{
		printf("\n TENLAB Error: Assignment dimension does not match.");
		TENLAB_Terminate();
	}
}


void TENLAB_Tensor_create(TENLAB_Tensor *X)
{
    X->Shape = NULL;
    X->Content = NULL;
    X->shape_length = 1;
	int *new_Shape = (int*) realloc(X->Shape, (X->shape_length) * sizeof(int));
	if(new_Shape != NULL)
    {
        X->Shape = new_Shape;
		X->Shape[0] = 1;
    }
	else
	{
		printf("\n TENLAB Error: Memory filled during tensor initialization.");
		TENLAB_Terminate();
	} 
	X->cur_content_length = 0;
	X->max_content_length = 0;
}

void TENLAB_Tensor_destroy(TENLAB_Tensor *X)
{
    free(X->Content);
	free(X->Shape);
    X->max_content_length = 0;
    X->cur_content_length = 0;
	X->shape_length = 0;
}

void TENLAB_Tensor_duplicate(TENLAB_Tensor *Y,TENLAB_Tensor *X)
{
    free(Y->Content);
	free(Y->Shape);
    Y->shape_length = X->shape_length;
	Y->cur_content_length = X->cur_content_length;
	Y->max_content_length = X->max_content_length;
	double *new_Content = (double*) malloc(X->max_content_length * sizeof(double));
	if(new_Content != NULL)
	{
		Y->Content = new_Content;
		memcpy(Y->Content,X->Content,X->max_content_length * sizeof(double));
	}
	else
	{
		printf("\n TENLAB Error: Memory filled during tensor duplication.");
		TENLAB_Terminate();
	} 
	int *new_Shape = (int*) malloc( X->shape_length * sizeof(int));
	if(new_Shape != NULL)
	{
		Y->Shape = new_Shape;
		memcpy(Y->Shape,X->Shape,X->shape_length * sizeof(int));
	}
	else
	{
		printf("\n TENLAB Error: Memory filled during tensor duplication.");
		TENLAB_Terminate();
	} 
}

void TENLAB_assign(TENLAB_Tensor *Y,TENLAB_Tensor X)
{
	TENLAB_Tensor temp;
	TENLAB_Tensor_create(&temp);
    Y->shape_length = X.shape_length;
	Y->cur_content_length = X.cur_content_length;
	Y->max_content_length = X.max_content_length;
	double *new_Content = (double*) malloc(X.max_content_length * sizeof(double));
	if(new_Content != NULL)
	{
		Y->Content = new_Content;
		memcpy(Y->Content,X.Content,X.max_content_length * sizeof(double));
	}
	else
	{
		printf("\n TENLAB Error: Memory filled during tensor duplication.");
		TENLAB_Terminate();
	} 
	int *new_Shape = (int*) malloc( X.shape_length * sizeof(int));
	if(new_Shape != NULL)
	{
		Y->Shape = new_Shape;
		memcpy(Y->Shape,X.Shape,X.shape_length * sizeof(int));
	}
	else
	{
		printf("\n TENLAB Error: Memory filled during tensor duplication.");
		TENLAB_Terminate();
	} 
	TENLAB_Tensor_destroy(&temp);
}

//#define TENLAB_assign(a, b) _Generic(b, int: TENLAB_add_element, double: TENLAB_add_element, TENLAB_Tensor : TENLAB_Tensor_nonpointing_duplicate)(a, b)

void TENLAB_Tensor_check_size(TENLAB_Tensor *X,TENLAB_Tensor *Y)
{
	if (X->Shape[0]!=Y->Shape[0])
	{
		printf("\n TENLAB Error: Total number of dimensions are different.");
		TENLAB_Terminate();
	}
	else
	{
		for (int i=1;i<=X->shape_length;i++)
		{
			if (X->Shape[i-1]!=X->Shape[i-1])
			{
				printf("\n TENLAB Error: Dimensions don't match.");
				TENLAB_Terminate();
			}
		}
	}
}

void TENLAB_Tensor_force_scalar(TENLAB_Tensor *X)
{
	if (X->cur_content_length>1)
	{
		printf("\n TENLAB Error: A scalar was sought.");
		TENLAB_Terminate();		
	}
}

void TENLAB_Tensor_while_is_not_scalar(TENLAB_Tensor *X)
{
	if (X->cur_content_length>1)
	{
		printf("\n TENLAB Warning: A while loop only considers the first element of a tensor.");
	}
}

void TENLAB_Tensor_print(TENLAB_Tensor X)
{
	if (X.cur_content_length>=1)
	{
		for(int i = 1; i <= X.cur_content_length; i++)
		{
			printf("%f\n",X.Content[i-1]);
		}
	}
}

void TENLAB_Tensor_shape_print(TENLAB_Tensor X)
{
    for(int i = 1; i <= X.shape_length; i++)
    {
        printf("%d\n",X.Shape[i-1]);
    }
}

void TENLAB_Tensor_populate_from_MEX(const mxArray *prhs[], TENLAB_Tensor *X, int idx)
{
    int TENLAB_elements;
    int TENLAB_dimensions;
    
    TENLAB_elements = mxGetNumberOfElements(prhs[idx]);
    TENLAB_dimensions = mxGetNumberOfDimensions(prhs[idx]);
    X->shape_length = TENLAB_dimensions;
	X->cur_content_length = TENLAB_elements;
	X->max_content_length = TENLAB_elements;
    double *el_pointer;
    el_pointer=(double *) mxGetPr(prhs[idx]);
	double *new_Content = (double*) malloc(X->max_content_length * sizeof(double));
	if(new_Content != NULL)
	{
        X->Content = new_Content;
        for(int j=1;j<=TENLAB_elements;j++)
        {
            X->Content[j-1] = el_pointer[j-1];
        }
	}
	else
	{
		printf("\n TENLAB Error: Memory filled during MEX copy.");
		TENLAB_Terminate();
	}
    int *new_Shape = (int *) malloc(X->shape_length * sizeof(int));
	if(new_Shape != NULL)
	{
        X->Shape = new_Shape;
        for(int j=1;j<=TENLAB_dimensions;j++)
        {
            X->Shape[j-1] = ((mxGetDimensions(prhs[idx]))[j-1]);
        }
	}
	else
	{
		printf("\n TENLAB Error: Memory filled during MEX copy.");
		TENLAB_Terminate();
	}
}

void TENLAB_Tensor_to_MEX(mxArray *plhs[], TENLAB_Tensor *X, int idx)
{
    mwSize *shape_out = (mwSize*) malloc( X->shape_length * sizeof(mwSize));
	if(shape_out != NULL)
	{
		for (int i=1;i<=(X->shape_length);i++)
        {
            shape_out[i-1]=X->Shape[i-1];
        }
	}
    else
	{
		printf("\n TENLAB Error: Memory filled during MEX conversion.");
		TENLAB_Terminate();
	} 
    plhs[idx] = mxCreateNumericArray(X->shape_length,shape_out,mxDOUBLE_CLASS,mxREAL);
    double *pr;
    pr = mxGetPr(plhs[idx]);
    for (int i=1;i<=(X->cur_content_length);i++)
    {
        pr[i-1] = X->Content[i-1];
    }
}